#------------------------------------------------------------------
# $Copyright: Copyright (c) 2022 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $Header: /project/isv-agent/VProAgents/PostgreSQL/PostgreSQL.pm,v 1.15 2015/03/04 13:09:54 mrampura Exp $
#
# $sAuthor: $
#
# $Date: 2015/03/04 13:09:54 $
# $Revision: 1.15 $
#
# Provides:                            PostgreSQL
# Author:                              ssrivast
# crAgent Build:                       1.91 2009/09/07 08:42:27
# crAgent Run Date:                    11 May, 2011
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package PostgreSQL;


#-------------------------------------
# Include required framework modules..
#-------------------------------------
use SystemD;        # Support generalized SystemD Functions
use PGReplication;  # Support PostgreSQL Replication Functions
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses
use Switch;
use ag_i18n_inc;
use POSIX qw ( uname );

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use Socket qw(:addrinfo);
use File::Path qw ( mkpath );
use File::Copy;
use File::stat;


my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my $sCmdHagrp  = $HAGRP  ||= "$CLUS_HOME/bin/hagrp";
my $sCmdHasys  = $HASYS  ||= "$CLUS_HOME/bin/hasys";
my $sCmdSSH = "/usr/bin/ssh";
my $sCmdCP = "/usr/bin/cp";
my $sCmdTouch = "/usr/bin/touch";

my @liRunningProcs = (); # Global..

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: PostgreSQL.pm,v 1.15 2015/03/04 13:09:54 mrampura Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define PostgreSQL subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String PostgreSQLUser,
                              IN String EnvFile,
                              IN String HostName,
                              IN Integer Port,
                              IN String BaseDir,
                              IN String DataDir,
                              IN String StartOpts,
                              IN String DBUser,
                              IN String DBName,
                              IN String Table,
                              IN Boolean UseSystemD,
                              IN String ServiceName,
                              IN Integer SecondLevelMonitor,
                              IN String MonitorProgram,
                              IN Boolean MonitorReplication,
                              IN String ClientAddr,
                              IN String SourceConnStr,
                              IN String BackupCmd,
                              IN String RecoveryFile
                              IN Boolean RestartdbToRewind );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $localEnvFile                  Absolute path to environment file to source
 $localHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $localPort                     The port number on which PostgreSQL instance is configured.
 $localBaseDir                  The Installation Directory of PostgreSQL.
 $localDataDir                  Location of the database storage area.
 $localStartOpts                Additional "postgres" option while starting the PostgreSQL instance.
 $localDBUser                   A valid dbuser name used to query during detail monitoring. This user should have query/update privileges to the Table in provided database.
 $localDBName                   A valid database name in which the Table is created for detailsed monitoring.
 $localTable                    A valid database table in the $DBUser schema on which the query is executed during detail monitoring. The Table should contain a single field "TSTAMP" with datatype DATE
 $localUseSystemD               When enabled, the reource comes online in system.slice
 $localServiceName              Name of the Service to start and stop PostgreSQL
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 $localMonitorProgram           Full path to external monitor program
 $localMonitorReplication       When enabled, Replication link is monitored. Takeover operation is enabled. Instance is connected as standby if Master is running.
 $localClientAddr               Address of the Standby instance
 $localSourceConnStr            Connection string used by pg_rewind command
 $localBackupCmd                Command used to base backup PostgreSQL data directory
 $localRecoveryFile             recovery.conf file provided to start standby instance
 $localRestartdbToRewind        When enabled, Database is shut down cleanly if rewind fails
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localPostgreSQLUser,
       $localEnvFile,
       $localHostName,
       $localPort,
       $localBaseDir,
       $localDataDir,
       $localStartOpts,
       $localDBUser,
       $localDBName,
       $localTable,
       $localUseSystemD,
       $localServiceName,
       $localSecondLevelMonitor,
       $localMonitorProgram,
       $localMonitorReplication,
       $localClientAddr,
       $localSourceConnStr,
       $localBackupCmd,
       $localRecoveryFile,
       $localRestartdbToRewind ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  PostgreSQLUser           [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  Port                     [%s]
  BaseDir                  [%s]
  DataDir                  [%s]
  StartOpts                [%s]
  DBUser                   [%s]
  DBName                   [%s]
  Table                    [%s]
  UseSystemD               [%s]
  ServiceName              [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]
  MonitorReplication       [%s]
  ClientAddr               [%s]
  SourceConnStr            [%s]
  BackupCmd                [%s]
  RecoveryFile             [%s]
  RestartdbToRewind        [%s]",

  $sMyName,   $sEntryPointName,   $localPostgreSQLUser,    $localEnvFile,
  $localHostName,    $localPort,     $localBaseDir,     $localDataDir,
  $localStartOpts, $localDBUser, $localDBName, $localTable, $localUseSystemD,
  $localServiceName,   $localSecondLevelMonitor,   $localMonitorProgram,
  $localMonitorReplication,    $localClientAddr,    $localSourceConnStr, 
  $localBackupCmd,       $localRecoveryFile,    $localRestartdbToRewind );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName  PostgreSQLUser  EnvFile  HostName  Port   BaseDir
       DataDir  StartOpts DBUser  DBName   Table UseSystemD  ServiceName
       SecondLevelMonitor MonitorProgram MonitorReplication ClientAddr 
       SourceConnStr    BackupCmd    RecoveryFile    RestartdbToRewind);

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------

  my @lsOnlineArgs  = qw ( BaseDir DataDir PostgreSQLUser HostName Port );
  my @lsOfflineArgs = qw ( BaseDir DataDir PostgreSQLUser );
  my @lsMonitorArgs = qw ( BaseDir DataDir PostgreSQLUser );
  my @lsCleanArgs   = qw ( BaseDir DataDir PostgreSQLUser );
  my @lsImfRegisterArgs   = qw ( BaseDir DataDir PostgreSQLUser );

  if ( $localSecondLevelMonitor ge 1 ) {
    @lsOnlineArgs  = qw ( BaseDir DataDir PostgreSQLUser HostName Port EnvFile DBUser DBName Table);
    @lsOfflineArgs = qw ( BaseDir DataDir PostgreSQLUser EnvFile DBUser DBName Table);
    @lsMonitorArgs = qw ( BaseDir DataDir PostgreSQLUser EnvFile DBUser DBName Table);
  }

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs,
                    'imf_register'	=> \@lsImfRegisterArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  
  #----------------------------------
  # Validation for all entry points..
  #----------------------------------

  if ( $sEntryPointName =~ /^(online|offline)$/o ) {
  #---------------------------
  # Validate parameter BaseDir..
  #---------------------------
    if ( ! -d $localBaseDir ) {
     &logError ( 20004,
        "<%s> BaseDir [%s] directory does not exist",
     $sMyName, $localBaseDir );
     $iReturnCode = 0; # Failure..
    }

    #---------------------------
    # Validate parameter DataDir..
    #---------------------------
    if ( ! -d $localDataDir ) {
      &logError ( 20005,
        "<%s> DataDir [%s] directory does not exist",
      $sMyName, $localDataDir );
      $iReturnCode = 0; # Failure..
    }

    if ($localMonitorReplication ){
      my @lsMonitorReplArgs = qw ( BaseDir DataDir PostgreSQLUser HostName Port 
                                   SourceConnStr BackupCmd RecoveryFile RestartdbToRewind);
      if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsMonitorReplArgs ) {
        &logError ( 20326,
        "<%s> The following mandatory fields were not set [%s], Total [%s]",
        $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
        $iReturnCode = 0; # Failure..
      }
      if (! -e $localRecoveryFile) {
        &logError ( 20449,
         "<%s> Recovery File [%s] does not exist.",
           $sMyName, $localRecoveryFile );
        $iReturnCode = 0;
      }
      my ($sCommand, $sCommandArgs) = $localBackupCmd
                                    =~ /(\S+)\s+(.*)/; 
      if ( ! -e $sCommand ) {
        &logError ( 20450, "<%s> Failed:Cmd [%s] does not exist.",
            $sMyName, $sCommand);
        $iReturnCode = 0;
      }
      elsif ( ! -x $sCommand ) {
          &logError ( 20451, "<%s> Failed:Cmd [%s] is not executable.",
            $sMyName, $sCommand);
        $iReturnCode = 0;
      }
      if ($localSourceConnStr !~ m/host/){
          &logError ( 20468, "<%s> SourceConnStr[%s] must contain the parameter host. ",
            $sMyName, $localSourceConnStr);
        $iReturnCode = 0;
      }
    }
  }

  #----------------------------------
  # Validate parameter PostgreSQLUser..
  #----------------------------------
  if ( ! defined getpwnam $localPostgreSQLUser ) {
     &logError ( 20006,
      "<%s> PostgreSQLUser [%s] is not a valid UNIX login",
     $sMyName, $localPostgreSQLUser );
     $iReturnCode = 0; # Failure..
  }
  #--------------------------------
  # Validate parameter ServiceName
  #--------------------------------
  
  if ( $localUseSystemD && ( $localServiceName !~ /\S/ ) ) {
    &logError ( 20322,
      "<%s> ServiceName attribute cannot be blank if UseSystemD attribute is set.",
      $sMyName );
    $iReturnCode = 0; #Failure
  }
 

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #------------------------------
    # Validate parameter HostName..
    #------------------------------
    my ($addr_err, @res) = getaddrinfo($localHostName);
    if ( ! @res ) {
      &logError ( 20007,
      "<%s> HostName [%s] is an invalid Hostname/IPv4/IPv6 address",
      $sMyName, $localHostName );
      $iReturnCode = 0; # Failure..
    }

    #--------------------------
    # Validate parameter Port..
    #--------------------------
    if ( ! $localPort ) {
      &logError ( 20008,
      "<%s> Port [%s] has to be greater than 0.",
      $sMyName, $localPort );
      $iReturnCode = 0; # Failure..
    }

    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20009,
      "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }
  }

  $iReturnCode ||
  &logWarn ( 20010,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String BaseDir,
                         IN String DBName,
                         IN String DBUser,
                         IN String DataDir,
                         IN String EnvFile,
                         IN String HostName,
                         IN Integer Port,
                         IN String PostgreSQLUser,
                         IN Integer SecondLevelMonitor,
                         IN String StartOpts,
                         IN String StopOpts,
                         IN String Table,
                         IN Boolean UseSystemD,
                         IN String ServiceName,
                         IN Boolean MonitorReplication,
                         IN String ClientAddr,
                         IN String SourceConnStr,
                         IN String BackupCmd,
                         IN String RecoveryFile,
                         IN Boolean RestartdbToRewind,
                         IN Integer RewindAttempt );

 Description : Start logic needed for online entry pt

 Parameters :

 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sDBName                   A valid database name in which the Table is created for detailsed monitoring.
 $sDBUser                   A valid dbuser name used to query during detail monitoring. This user should have query/update privileges to the Table in provided database.
 $sDataDir                  Location of the database storage area.
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $iPort                     The port number on which PostgreSQL instance is configured.
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 $sStartOpts                Additional "postgres" option while starting the PostgreSQL instance.
 $sStopOpts                 Additional "postgres" option while stopping the PostgreSQL instance.
 $sTable                    A valid database table in the $DBUser schema on which the query is executed during detail monitoring. The Table should contain a single field "TSTAMP" with datatype DATE
 $bUseSystemD               When enabled, the reource comes online in system.slice
 $sServiceName              Name of the Service to start and stop PostgreSQL
 $bMonitorReplication       When enabled, Replication link is monitored. Takeover operation is enabled. Instance is connected as standby if Master is running.
 $sClientAddr               Address of the Standby instance
 $sSourceConnStr            Connection string used by pg_rewind command
 $sBackupCmd                Command used to base backup PostgreSQL data directory
 $sRecoveryFile             recovery.conf file provided to start standby instance
 $bRestartdbToRewind        When enabled, Database is shut down cleanly if rewind fails
 $iRewindAttempt            Number of Attempts made to backup using pg_rewind command

 
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDBName,
       $sDBUser,
       $sDataDir,
       $sEnvFile,
       $sHostName,
       $iPort,
       $sPostgreSQLUser,
       $iSecondLevelMonitor,
       $sStartOpts,
       $sStopOpts,
       $sTable,
       $bUseSystemD,
       $sServiceName,
       $bMonitorReplication,
       $sClientAddr,
       $sSourceConnStr,
       $sBackupCmd,
       $sRecoveryFile,
       $bRestartdbToRewind,
       $iRewindAttempt) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20011, "Subroutine <%s> called with:
  BaseDir                  [%s]
  DBName                   [%s]
  DBUser                   [%s]
  DataDir                  [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  Port                     [%s]
  PostgreSQLUser           [%s]
  SecondLevelMonitor       [%s]
  StartOpts                [%s]
  StopOpts                 [%s]
  Table                    [%s]
  UseSystemD               [%s]
  ServiceName              [%s]
  MonitorReplication       [%s]
  ClientAddr               [%s]
  SourceConnStr            [%s]
  BackupCmd                [%s]
  RecoveryFile             [%s]
  RestartdbToRewind        [%s]
  RewindAttempt            [%s]",

  $sMyName, $sBaseDir, $sDBName, $sDBUser, $sDataDir, $sEnvFile,
  $sHostName, $iPort, $sPostgreSQLUser, $iSecondLevelMonitor,
  $sStartOpts, $sStopOpts, $sTable, $bUseSystemD, $sServiceName,
  $bMonitorReplication, $sClientAddr, $sSourceConnStr, $sBackupCmd,
  $sRecoveryFile, $bRestartdbToRewind, $iRewindAttempt );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iRet, $iTimeout,
       @lsCmdout, @liPids, $sReplicationModeTuples ) = ();

  #-----------------------------------------
  # Whack any preexisting PostgreSQL processes..
  #-----------------------------------------
  &logTrace ( 20012,
  "<%s> Killing PostgreSQL processes for this instance", $sMyName );

  if ( PostgreSQL::KillPostgreSQLProcesses ( $sBaseDir,
                                             $sDataDir,
                                             $sEnvFile,
                                             $sHostName,
                                             $iPort,
                                             $sPostgreSQLUser,
                                             $bMonitorReplication,
                                             $sClientAddr ) ) {
    &logTrace ( 20013,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20014,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    PostgreSQL::DeletePidFile ( );
    $iReturnCode = 0; # Failure..
  }
  
  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + PostgreSQL::GetEngineTimeout ( 1 );
  
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };

  #---------------------------------------
  # backup if need to set up as standby.
  #---------------------------------------
  
  my $sMaster = ();
  my $bRewindFailed = 0;
  if ($bMonitorReplication == 1 && $iRewindAttempt < 2) {
    $iRet = PostgreSQL::IsMasterRunning( $sBaseDir,
                                         $sPostgreSQLUser,
                                         $sEnvFile,
                                         $sSourceConnStr,
                                         $iTillWhen,
                                         \$sMaster,                                                                            
                                         $iPort ) ; 
    if ( $iRet == 1) {
      $iRet = PostgreSQL::ConnectAsStandby( $sBaseDir, 
                                            $sDataDir, 
                                            $iPort, 
                                            $sHostName, 
                                            $sPostgreSQLUser, 
                                            $sSourceConnStr, 
                                            $sBackupCmd,  
                                            $sRecoveryFile,
                                            $sEnvFile, 
                                            $iTillWhen,
                                            $sMaster );
      if ($iRet == 1){          # Success. Either with basebackup or rewind.
        &logTrace ( 20353,
                   "<%s> Successfully configured this database instance to connect as standby.",$sMyName );
      }
      elsif ($iRet == -1) {     # Rewind failed. 
        if (($iRewindAttempt == 1) || ($bRestartdbToRewind == 0)){ 
# Return failure. (rewind failed second time OR RestartdbToRewind attribute not set.)
          &logError ( 20354,
                   "<%s> Rewind has failed. Cannot connect as standby. Manually configure the instance to connect as standby and online again.",$sMyName );
          return 0;
        }
        $bRewindFailed = 1;
      }  
      elsif ( $iRet == 2 ) {
       # Failed to find master. Infoscale commands failed.
        &logError ( 20647,
                   "<%s> Failed to find master. Master May or Maynot be running. Manually configure the instance to connect as standby and online again.",$sMyName ); 
        return 0; # Failure..
      }
      else {
# Failure to connect as standby. Either basebackup failed or other postgres commands failed.
        &logError ( 20355,
                   "<%s> Failed to connect as standby. Manually configure the instance to connect as standby and online again.",$sMyName ); 
        return 0; # Failure..
      }
    }
    elsif ( $iRet == 2 ) {
    # Failure to find master. Infoscale commands failed.
        &logError ( 20647,
                   "<%s> Failed to find master. Master May or Maynot be running.Manually configure the instance to connect as standby and online again.",$sMyName ); 
        return 0; # Failure..
    }
    else {
    # continue. No action required
     &logTrace ( 20356,
                  "<%s> No Master instance running. Will attempt to start as master.",$sMyName );  
    }
  }

  if ( PostgreSQL::GetAttrValue("ReplicationModeTuples", $sResourceName,"", \$sReplicationModeTuples ) && defined( $sReplicationModeTuples ) && $sReplicationModeTuples ne "" ) {
    if ( ( $bRewindFailed == 1 ) && ( $iRewindAttempt == 0 ) && ( $bRestartdbToRewind == 1 ) ) {
   
      $iRewindAttempt = $iRewindAttempt + 1 ; # Start again. We have already attempted rewind once. This is second time
      if ( PostgreSQL::Start ( $sBaseDir,
                               $sDBName,
                               $sDBUser,
                               $sDataDir,
                               $sEnvFile,
                               $sHostName,
                               $iPort,
                               $sPostgreSQLUser,
                               $iSecondLevelMonitor,
                               $sStartOpts,
                               $sStopOpts,
                               $sTable,
                               $bUseSystemD,
                               $sServiceName,
                               $bMonitorReplication,
                               $sClientAddr,
                               $sSourceConnStr,
                               $sBackupCmd,
                               $sRecoveryFile,
                               $bRestartdbToRewind,
                               $iRewindAttempt ) ) {
        &logTrace ( 20358, "<%s> Resource [%s] started up successfully.",
        $sMyName, $sResourceName );
      }
      else {
	return 0; #Failure ( second rewind failed )
      }
    }
  }

  $iRet = $bUseSystemD && SystemD::DetectSystemd();
  if(!$iRet) {
    #------------------------------------------
    # Define start command and args variables..
    #------------------------------------------
    $sCommand = "$sBaseDir/pg_ctl";

    $sCommandArgs = "start -w -D $sDataDir -o \"-p $iPort -h $sHostName \" $sStartOpts &";
    if ( $bMonitorReplication == 1 ) {
      $sCommandArgs = "start -w -D $sDataDir -o \"-p $iPort \" $sStartOpts &";
    }

    ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  
    &logTrace ( 20015,
    "<%s> Starting PostgreSQL instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  } 
  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20016,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ($iRet) {
    if( $iReturnCode = SystemD::RunSystemdServiceName($sResourceName, $sServiceName, "start" )){
        &logTrace ( 20323,
        "<%s> RunSystemd was successful",
        $sMyName );
    }
  }
  elsif ( $iReturnCode =
    PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser, 
                               $sEnvFile, 
                               $sCommand, 
                               $sCommandArgs, 
                               \$iExitCode, 
                               $iTimeout, 
                               \@lsCmdout ) ) {
    &logTrace ( 20017,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;


  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20018,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         PostgreSQL::GetProcState ( \@liPids,
                                    $sBaseDir,
                                    $sDataDir,
                                    $sEnvFile,
                                    $sHostName,
                                    $iPort,
                                    $sPostgreSQLUser,
                                    $bMonitorReplication,
                                    $sClientAddr,
                                    1 ) && (           # Running FFLM..
         $iSecondLevelMonitor  ?                              # If enabled..
         VCSDef_MonitorExit::ONLINE ==                        # Online after..
         PostgreSQL::SecondLevelMonitor ( $sBaseDir,   # Running SLM..
                                          $sDBName,
                                          $sDBUser,
                                          $sEnvFile,
                                          $sHostName,
                                          $iPort,
                                          $sPostgreSQLUser,
                                          $sTable,
                                          $iTillWhen ) : 1 ) ) { 
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20019,
      "<%s> Confirmed start-up of this instance of PostgreSQL server",
      $sMyName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20020,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0; # Failure..
    }
  }

  #--------------------------------------------------------
  # if rewind had failed once stop the database gracefully.
  # Rewind and attempt to online again.
  #--------------------------------------------------------

  if (($bRewindFailed == 1) && ($iRewindAttempt == 0) && ( $bRestartdbToRewind == 1 )){
   
    &logTrace ( 20357,
      "<%s> Attempting to shutdown cleanly.",
      $sMyName );
    if ( PostgreSQL::Stop ( $sBaseDir,
                            $sDataDir,
                            $sEnvFile,
                            $sHostName,
                            $iPort,
                            $sPostgreSQLUser,
                            $sStopOpts,
                            $bUseSystemD,
                            $sServiceName,
                            $bMonitorReplication,
                            $sClientAddr ) ) {
      &logInfo ( 20467,
      "<%s> Shutdown resource [%s] successfully.",
      $sMyName, $sResourceName );
      $iRewindAttempt = 1 ; # Start again. We have already attempted rewind once. This is second time
      if ( PostgreSQL::Start ( $sBaseDir,
                               $sDBName,
                               $sDBUser,
                               $sDataDir,
                               $sEnvFile,
                               $sHostName,
                               $iPort,
                               $sPostgreSQLUser,
                               $iSecondLevelMonitor,
                               $sStartOpts,
                               $sStopOpts,
                               $sTable,
                               $bUseSystemD,
                               $sServiceName,
                               $bMonitorReplication,
                               $sClientAddr,
                               $sSourceConnStr,
                               $sBackupCmd,
                               $sRecoveryFile,
                               $bRestartdbToRewind,
                               $iRewindAttempt ) ) {
        &logTrace ( 20358, "<%s> Resource [%s] started up successfully.",
        $sMyName, $sResourceName );
      }
      else {
        &logError( 20359, "<%s> Failed to online resource", $sMyName );
      }
    }
    else {
      &logError ( 20360, "<%s> Failed to shutdown the database cleanly", $sMyName );
    }
  }
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String BaseDir,
                        IN String DataDir,
                        IN String EnvFile,
                        IN String HostName,
                        IN String PostgreSQLUser,
                        IN String StopOpts,
                        IN Boolean UseSystemD,
                        IN String ServiceName,
                        IN Boolean MonitorReplication,
                        IN String ClientAddr );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sDataDir                  Location of the database storage area.
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $sStopOpts                 Additional "postgres" option while stopping the PostgreSQL instance.
 $bUseSystemD               When enabled, the reource comes online in system.slice
 $sServiceName              Name of the Service to start and stop PostgreSQL
 $bMonitorReplication       When enabled, Replication link is monitored. Takeover operation is enabled. Instance is connected as standby if Master is running.
 $sClientAddr               Address of the Standby instance


 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDataDir,
       $sEnvFile,
       $sHostName,
       $iPort,
       $sPostgreSQLUser,
       $sStopOpts,
       $bUseSystemD,
       $sServiceName,
       $bMonitorReplication,
       $sClientAddr ) = @_;


  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20021, "Subroutine <%s> called with:
  BaseDir                  [%s]
  DataDir                  [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  PostgreSQLUser           [%s]
  StopOpts                 [%s]
  UseSystemD               [%s]
  ServiceName              [%s]
  MonitorReplication       [%s]
  ClientAddr               [%s]",

  $sMyName,   $sBaseDir,    $sDataDir,    $sEnvFile,    $sHostName,
  $sPostgreSQLUser, $sStopOpts,  $bUseSystemD, $sServiceName,
  $bMonitorReplication, $sClientAddr );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + PostgreSQL::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
    my $iRet = $bUseSystemD && SystemD::DetectSystemd();

  if(!$iRet) {
    #------------------------------------
    # Define stop command and arguments..
    #------------------------------------
    $sCommand = "$sBaseDir/pg_ctl";
    $sCommandArgs = "stop -w -D $sDataDir";
    $sCommandArgs .= " $sStopOpts " if ($sStopOpts =~ /\S/);
  
    &logTrace ( 20022,
    "<%s> Stopping PostgreSQL with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  }
  ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate..
  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20023,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ($iRet) {
    my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
    if( $iReturnCode = SystemD::RunSystemdServiceName($sResourceName, $sServiceName, "stop" )){
        &logTrace ( 20324,
        "<%s> RunSystemd was successful",
        $sMyName );
    }
  }
  elsif ( $iReturnCode =
          PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser, 
                                     $sEnvFile, 
                                     $sCommand, 
                                     $sCommandArgs, 
                                     \$iExitCode, 
                                     $iTimeout, 
                                     \@lsCmdout ) ) {
    &logTrace ( 20024,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }


  #--------------------
  # Delay entry point..
  #--------------------
  if ( $iReturnCode ) { # Success..
    #-----------------
    # Set sleep time..
    #-----------------
    $iSleepTime = 10; # Sec..
  
    &logInfo ( 20025,
    "<%s> Delaying exit to confirm shutdown..", $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
      #-----------------------
      # Wait before retrying..
      #-----------------------
      sleep $iSleepTime; # Secs..
  
      PostgreSQL::GetProcState ( \@liPids,
                                 $sBaseDir,
                                 $sDataDir,
                                 $sEnvFile,
                                 $sHostName,
                                 $iPort,
                                 $sPostgreSQLUser,
                                 $bMonitorReplication,
                                 $sClientAddr,
                                 0 ); # Request regular..
  
      #------------------------------
      # Check for pending processes..
      #------------------------------
      if ( scalar @liPids ) {
        &logTrace ( 20026,
        "<%s> Could not confirm shutdown. Retrying in [%s] sec.
        Delaying kill with [%s] sec in hand",
        $sMyName, $iSleepTime, $iTimeout - $iSleepTime );
      }
      else {
        &logInfo ( 20027,
        "<%s> Confirmed shutdown of PostgreSQL instance",
        $sMyName );
        last; # Get out..
      }
    }
  }

  #-----------------------------------------
  # Whack any persisting PostgreSQL processes..
  #-----------------------------------------
  &logTrace ( 20028,
  "<%s> Killing PostgreSQL processes for this instance", $sMyName );

  if ( PostgreSQL::KillPostgreSQLProcesses ( $sBaseDir,
                                             $sDataDir,
                                             $sEnvFile,
                                             $sHostName,
                                             $iPort,
                                             $sPostgreSQLUser,
                                             $bMonitorReplication,
                                             $sClientAddr ) ) {
    &logTrace ( 20029,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20030,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    PostgreSQL::DeletePidFile ( );
    $iReturnCode = 0; # Failure..
  }
  
  

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN String BaseDir,
                        IN Integer CleanReason,
                        IN String DataDir,
                        IN String EnvFile,
                        IN String HostName,
                        IN String PostgreSQLUser,
                        IN String StopOpts,
                        IN Boolean UseSystemD,
                        IN String ServiceName 
                        IN Boolean MonitorReplication,
                        IN String ClientAddr )

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $sBaseDir                  The Installation Directory of PostgreSQL.
 $iCleanReason              Why clean was called
 $sDataDir                  Location of the database storage area.
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $sStopOpts                 Additional "postgres" option while stopping the PostgreSQL instance.
 $bUseSystemD               When enabled, the reource comes online in system.slice
 $sServiceName              Name of the Service to start and stop PostgreSQL
 $bMonitorReplication       When enabled, Replication link is monitored. Takeover operation is enabled. Instance is connected as standby if Master is running.
 $sClientAddr               Address of the Standby instance
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $iCleanReason,
       $sDataDir,
       $sEnvFile,
       $sHostName,
       $iPort,
       $sPostgreSQLUser,
       $sStopOpts,
       $bUseSystemD,
       $sServiceName,
       $bMonitorReplication,
       $sClientAddr,
       $sRecoveryFile ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20031, "Subroutine <%s> called with:
  BaseDir                  [%s]
  CleanReason              [%s]
  DataDir                  [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  PostgreSQLUser           [%s]
  StopOpts                 [%s]
  UseSystemD               [%s]
  ServiceName              [%s]
  MonitorReplication       [%s]
  ClientAddr               [%s]
  RecoveryFile             [%s]",

  $sMyName,   $sBaseDir,   $sCleanReason,   $sDataDir,   $sEnvFile,
  $sHostName, $sPostgreSQLUser, $sStopOpts, $bUseSystemD, $sServiceName,
  $bMonitorReplication, $sClientAddr, $sRecoveryFile );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + PostgreSQL::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsFiles, @lsCmdout ) = ();
  
  #-------------------------------------
  # Define clean command and arguments..
  #-------------------------------------
  $sCommand = "$sBaseDir/pg_ctl";
  $sCommandArgs = "stop -w -D $sDataDir";
  $sCommandArgs .= " $sStopOpts " if ($sStopOpts =~ /\S/);

  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
  my $iRet = $bUseSystemD && SystemD::DetectSystemd();

  if(!$iRet) {
    &logInfo ( 20032,
    "<%s> Attempting to stop PostgreSQL with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  }
    #------------------------
    # Request shutdown..
    #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
     &logError ( 20033,
     "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
     $sMyName, abs ( $iTimeout ) );

     $iReturnCode = 0; # Failure..
   }
   elsif ($iRet) {
   my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
   if( $iReturnCode = SystemD::RunSystemdServiceName($sResourceName, $sServiceName, "stop" )){
       &logTrace ( 20325,
       "<%s> RunSystemd was successful",
       $sMyName );
   }
 }
 elsif ( PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser, 
                                    $sEnvFile, 
                                    $sCommand, 
                                    $sCommandArgs, 
                                    \$iExitCode, 
                                    $iTimeout, 
                                    \@lsCmdout ) ) {
    &logTrace ( 20034, 
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  }

  #-----------------------------------------
  # Whack any persisting PostgreSQL processes..
  #-----------------------------------------
  &logTrace ( 20035,
  "<%s> Killing PostgreSQL processes for this instance", $sMyName );

  if ( PostgreSQL::KillPostgreSQLProcesses ( $sBaseDir,
                                             $sDataDir,
                                             $sEnvFile,
                                             $sHostName,
                                             $iPort,
                                             $sPostgreSQLUser,
                                             $bMonitorReplication,
                                             $sClientAddr ) ) {
    &logTrace ( 20036,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20037,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    PostgreSQL::DeletePidFile ( );
    $iReturnCode = 0; # Failure..
  }
 ( my $sFile = $sRecoveryFile ) =~ s#.*/(.*)$#$1#;
  Sys::UnlinkFilesAsUser ( "$sDataDir/$sFile" ); 
  
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String BaseDir,
                                   IN String DataDir,
                                   IN String HostName,
                                   IN String PostgreSQLUser,
                                   IN Boolean MonitorReplication,
                                   IN String ClientAddr );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sDataDir                  Location of the database storage area.
 $sHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $bMonitorReplication       When enabled, Replication link is monitored. Takeover operation is enabled. Instance is connected as standby if Master is running.
 $sClientAddr               Address of the Standby instance
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDataDir,
       $sEnvFile,
       $sHostName,
       $iPort,
       $sPostgreSQLUser,
       $bMonitorReplication,
       $sClientAddr ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20038, "Subroutine <%s> called:
  BaseDir                  [%s]
  DataDir                  [%s]
  HostName                 [%s]
  PostgreSQLUser           [%s]
  MonitorReplication       [%s]
  ClientAddr               [%s]",

  $sMyName, $sBaseDir, $sDataDir, $sHostName, $sPostgreSQLUser,
  $bMonitorReplication, $sClientAddr );
 
  my @liDummyPids = ();
  my $iReturnCode = PostgreSQL::GetProcState ( \@liDummyPids,
                                               $sBaseDir,
                                               $sDataDir,
                                               $sEnvFile,
                                               $sHostName,
                                               $iPort,
                                               $sPostgreSQLUser,
                                               $bMonitorReplication,
                                               $sClientAddr,
                                               1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20040,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20041,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;
                                                                                
  &logTrace ( 20042,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String BaseDir,
                               IN String DataDir,
                               IN String HostName,
                               IN String PostgreSQLUser,
                               IN Boolean MonitorReplication,
                               IN String ClientAddr,
                               IN Integer UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sDataDir                  Location of the database storage area.
 $sHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $bMonitorReplication       When enabled, Replication link is monitored. Takeover operation is enabled. Instance is connected as standby if Master is running.
 $sClientAddr               Address of the Standby instance
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sBaseDir,
       $sDataDir,
       $sEnvFile,
       $sHostName,
       $iPort,
       $sPostgreSQLUser,
       $bMonitorReplication,
       $sClientAddr,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20043, "Subroutine <%s> called with:
  OnlinePids               [%s]
  BaseDir                  [%s]
  DataDir                  [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  Port                     [%s]
  PostgreSQLUser           [%s]
  MonitorReplication       [%s]
  ClientAddr               [%s]
  UseFFLM                  [%s]",

  $sMyName, $rliOnlinePids, $sBaseDir, $sDataDir, $sEnvFile, $sHostName,
  $iPort, $sPostgreSQLUser, $bMonitorReplication, $sClientAddr, $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsValues,
       %hiProcPids, %hsSrvrDetails, %hsCopy ) = ();

  @$rliOnlinePids = ();
  

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == ( $iUid = getpwnam $sPostgreSQLUser ) ) && # User is the super-user..
       ( 2 == $iUseFFLM ) ) {                  # And called to kill-by-UID..
    &logTrace ( 20044, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sPostgreSQLUser, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( PostgreSQL::GetSrvrDetails ( $sBaseDir, 
                                      $sDataDir, 
                                      $sHostName, 
                                      \%hsSrvrDetails ) ) {
      &logTrace ( 20045,
      "<%s> Got details for monitoring PostgreSQL instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20046,
      "<%s> Failed to get details for monitoring PostgreSQL instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nPostgreSQL process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = PostgreSQL::FFLMStatus ( $sPidFile,
                                         $sProc,
                                         $iUid,
                                         \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20047,
      "<%s> PostgreSQL instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20048,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sPostgreSQLUser, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20049,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20050,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20051,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

=pod
      &logTrace ( 20052, "<%s> Searching UID [%s] PID [%s] PPID [%s] CMD [%s]",
      $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
      $$rhlProc { COMMAND } );
=cut

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) =
                      splice ( @lsValues, 0, 3 ) ) {
              if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
                if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {    # Now found..
                  unshift ( @$rliOnlinePids, $iPid );         # Store..
    
                  &logTrace ( 20052,
                  "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                  $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                  $sProc );

                  $hiProcPids { $sKey } = $iPid;        # Store..
                  $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..
                }
              }
    
              #----------------
              # Update status..
              #----------------
              push ( @lsUpdates, $sPidFile, $sProc, $iFound );
            }
    
            @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
          }
        }
        else { # Request for kill..
          &logTrace ( 20053,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..
        } 
      }
      
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $iUseFFLM != 2                            &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20054,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( PostgreSQL::CreatePidFile ( $sPidFile,
                                     $hiProcPids { $sKey } ) ) {
          &logTrace ( 20055,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20056,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
          PostgreSQL::DeletePidFile ( );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20057,
    "<%s> Found [%s] PIDs [%s] for this instance of PostgreSQL.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }
  if ( $bMonitorReplication && $iReturnCode == VCSDef_MonitorExit::ONLINE){
    if (PostgreSQL::MonitorReplication($sBaseDir,
                                   $sEnvFile,
                                   $sHostName,
                                   $iPort,
                                   $sPostgreSQLUser,
                                   $sClientAddr  ) ) {
     &logTrace ( 20327,
    "<%s> Replication Monitoring successful.",
    $sMyName ); 
    }
    else  {
           &logWarn ( 20328,
    "<%s> Replication Monitoring failed.",
    $sMyName );
    }
  }
  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20058,
    "<%s> PostgreSQL process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String BaseDir,
                          IN String DBName,
                          IN String DBUser,
                          IN String EnvFile,
                          IN String HostName,
                          IN String MonitorProgram,
                          IN Integer Port,
                          IN String PostgreSQLUser,
                          IN Integer SecondLevelMonitor,
                          IN String Table );

 Description: Performs the in-depth monitoring

 Parameters :

 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sDBName                   A valid database name in which the Table is created for detailsed monitoring.
 $sDBUser                   A valid dbuser name used to query during detail monitoring. This user should have query/update privileges to the Table in provided database.
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $sMonitorProgram           Full path to external monitor program
 $iPort                     The port number on which PostgreSQL instance is configured.
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 $sTable                    A valid database table in the $DBUser schema on which the query is executed during detail monitoring. The Table should contain a single field "TSTAMP" with datatype DATE
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDBName,
       $sDBUser,
       $sEnvFile,
       $sHostName,
       $sMonitorProgram,
       $iPort,
       $sPostgreSQLUser,
       $iSecondLevelMonitor,
       $sTable,
       $level_two ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20059, "Subroutine <%s> called with:
  BaseDir                  [%s]
  DBName                   [%s]
  DBUser                   [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  MonitorProgram           [%s]
  Port                     [%s]
  PostgreSQLUser           [%s]
  SecondLevelMonitor       [%s]
  Table                    [%s]
  level_two                [%s]",

  $sMyName, $sBaseDir, $sDBName, $sDBUser,  $sEnvFile,  $sHostName,
  $sMonitorProgram, $iPort, $sPostgreSQLUser, $iSecondLevelMonitor,
  $sTable, $level_two );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #---------------------------------------
  # Do Second level monitoring if enabled.
  #---------------------------------------
  if ( -300 == $level_two ) {
     $level_two = VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor );  
     &logTrace ( 20307, "<%s> Traditional second level monitoring, SLM flag for this monitor set to [%s] ", $sMyName, $level_two  );
  }
  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( $level_two >= 1 ) 
  {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         PostgreSQL::DelayedAttributeCheck ( $sBaseDir,
                                             $sDBName,
                                             $sDBUser,
                                             $sEnvFile,
                                             $sHostName,
                                             $iPort,
                                             $sPostgreSQLUser,
                                             $sTable ) ) ) {
      &logTrace ( 20060,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20061,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20062,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + PostgreSQL::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         PostgreSQL::SecondLevelMonitor ( $sBaseDir,
                                          $sDBName,
                                          $sDBUser,
                                          $sEnvFile,
                                          $sHostName,
                                          $iPort,
                                          $sPostgreSQLUser,
                                          $sTable,
                                          $iTillWhen ) ) ) {

      &logTrace ( 20063,
      "<%s> PostgreSQL instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20064,
      "<%s> PostgreSQL instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + PostgreSQL::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         PostgreSQL::ExternalMonitor ( $sPostgreSQLUser,
                                       $sEnvFile,
                                       $sExtCmd,
                                       $sExtCmdArgs,
                                       $iTillWhen ) ) ) {
      &logTrace ( 20065,
      "<%s> PostgreSQL instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20066,
      "<%s> PostgreSQL instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String BaseDir,
                                        IN String DBName,
                                        IN String DBUser,
                                        IN String EnvFile,
                                        IN String HostName,
                                        IN Integer Port,
                                        IN String PostgreSQLUser,
                                        IN String Table );

 Description : Just In Time Validation

 Parameters :

 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sDBName                   A valid database name in which the Table is created for detailsed monitoring.
 $sDBUser                   A valid dbuser name used to query during detail monitoring. This user should have query/update privileges to the Table in provided database.
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $iPort                     The port number on which PostgreSQL instance is configured.
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $sTable                    A valid database table in the $DBUser schema on which the query is executed during detail monitoring. The Table should contain a single field "TSTAMP" with datatype DATE
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDBName,
       $sDBUser,
       $sEnvFile,
       $sHostName,
       $iPort,
       $sPostgreSQLUser,
       $sTable ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20067, "Subroutine <%s> called with:
  BaseDir                  [%s]
  DBName                   [%s]
  DBUser                   [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  Port                     [%s]
  PostgreSQLUser           [%s]
  Table                    [%s]",

  $sMyName, $sBaseDir, $sDBName, $sDBUser,  $sEnvFile,  $sHostName,
  $iPort, $sPostgreSQLUser, $sTable );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( BaseDir DBName DBUser EnvFile HostName Port PostgreSQLUser Table );
  my @lsReqdArgs = qw ( BaseDir DBName DBUser EnvFile HostName Port PostgreSQLUser Table );
  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20068,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  if ( ! -e $sEnvFile ) {
    &logError ( 20069,
    "<%s> EnvFile [%s] file does not exist",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #------------------------------
  # Validate parameter BaseDir..
  #------------------------------
  if ( ! -d $sBaseDir ) {
    &logError ( 20070,
    "<%s> BaseDir [%s] directory does not exist",
    $sMyName, $sBaseDir );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-------------------------------------
  # Validate parameter PostgreSQLUser..
  #-------------------------------------
  if ( ! defined getpwnam $sPostgreSQLUser ) {
    &logError ( 20071,
    "<%s> PostgreSQLUser [%s] is not a valid UNIX login",
    $sMyName, $sPostgreSQLUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-------------------------------
  # Validate parameter HostName..
  #-------------------------------
  my ($addr_err, @res) = getaddrinfo($sHostName);
  if ( ! @res ) {
    &logError ( 20072,
    "<%s> HostName [%s] is an invalid Hostname/IPv4/IPv6 address",
    $sMyName, $sHostName );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #---------------------------
  # Validate parameter Port..
  #---------------------------
  if ( ! $iPort ) {
    &logError ( 20073,
    "<%s> Port [%s] has to be greater than 0, if SLM is enabled",
    $sMyName, $iPort );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String BaseDir,
                                     IN String DBName,
                                     IN String DBUser,
                                     IN String EnvFile,
                                     IN String HostName,
                                     IN Integer Port,
                                     IN String PostgreSQLUser,
                                     IN String Table,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sDBName                   A valid database name in which the Table is created for detailsed monitoring.
 $sDBUser                   A valid dbuser name used to query during detail monitoring. This user should have query/update privileges to the Table in provided database.
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $iPort                     The port number on which PostgreSQL instance is configured.
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $sTable                    A valid database table in the $DBUser schema on which the query is executed during detail monitoring. The Table should contain a single field "TSTAMP" with datatype DATE
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDBName,
       $sDBUser,
       $sEnvFile,
       $sHostName,
       $iPort,
       $sPostgreSQLUser,
       $sTable,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20074, "Subroutine <%s> called with:
  BaseDir                  [%s]
  DBName                   [%s]
  DBUser                   [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  Port                     [%s]
  PostgreSQLUser           [%s]
  Table                    [%s]
  TillWhen                 [%s]",

  $sMyName, $sBaseDir, $sDBName, $sDBUser,  $sEnvFile,  $sHostName,
  $iPort, $sPostgreSQLUser, $sTable, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();
  
  #---------------------------------------------
  # Declare monitor command and args variables..
  #---------------------------------------------
  $sCommand = "$sBaseDir/psql";
  $sCommandArgs = "-A -t -h $sHostName -p $iPort -U $sDBUser -d $sDBName -c \"SELECT TSTAMP FROM $sTable\"";

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20075,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser, 
                                     $sEnvFile, 
                                     $sCommand, 
                                     $sCommandArgs, 
                                     \$iExitCode, 
                                     $iTimeout, 
                                     \@lsCmdout ) ) {
    &logTrace ( 20076,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
    
    if ( 0 == $iExitCode )
    {
		&logTrace ( 20077,
		"<%s> Select sql executed successfully with output [%s]",
		$sMyName, join ( chr 012, @lsCmdout ) );
    }
    else 
    {
		&logTrace ( 20078,
		"<%s> PSQL select statement failed with Exit Code [%s] and output [%s]", 
		$sMyName, $iExitCode, join ( chr 012, @lsCmdout ) );
        $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  elsif ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int MonitorReplication ( IN String BaseDir,
                                     IN String EnvFile,
                                     IN String HostName,
                                     IN Integer Port,
                                     IN String PostgreSQLUser,
                                     IN String ClientAddr );

 Description : Monitors Replication Link, if enabled.

 Parameters :

 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $iPort                     The port number on which PostgreSQL instance is configured.
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $sClientAddr               Address of the Standby instance
 
 Returns : boolean value in the set of {0,1} where
           0 false - Replication not active
           1 true  - Replication active
 -------------------------------------------------------------------------

=cut

sub MonitorReplication {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sEnvFile,
       $sHostName,
       $iPort,
       $sPostgreSQLUser,
       $sClientAddr ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20329, "Subroutine <%s> called with:
  BaseDir                  [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  Port                     [%s]
  PostgreSQLUser           [%s]
  ClientAddr               [%s]",

  $sMyName, $sBaseDir, $sEnvFile,  $sHostName,
  $iPort, $sPostgreSQLUser, $sClientAddr  );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();
  
  #---------------------------------------------
  # Declare monitor command and args variables..
  #---------------------------------------------
  $sCommand = "$sBaseDir/psql";
  $sCommandArgs = "-A -x -h $sHostName -p $iPort -c \"select pg_is_in_recovery\(\)\"";
  $iTimeout = 20;
    if ( PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser,
                                     $sEnvFile,
                                     $sCommand,
                                     $sCommandArgs,
                                     \$iExitCode,
                                     $iTimeout,
                                     \@lsCmdout ) ) {
    &logTrace ( 20330,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( 0 == $iExitCode )
    {
                &logTrace ( 20331,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
      chomp $lsCmdout[0];
      if ( grep(/pg_is_in_recovery\|t/,@lsCmdout) ) {
        &logTrace ( 20332,
                "<%s> This is standby server. ",
                $sMyName );
        return 1;
      }
      elsif ( grep(/pg_is_in_recovery\|f/,@lsCmdout) ) {
        &logTrace ( 20452,
                "<%s> This is master server. ",
                $sMyName );
      }
    }
    else
    {
      &logWarn ( 20333,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, join ( chr 012, @lsCmdout ) );
       return 0 # Failure..
    }
  }
  else
  {
    &logError ( 20334,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }

  #-------------------------------------------------------------
  # Sample Output:
  # 
  #  postgres@vcsesx001:~> /usr/pgsql-10/bin/psql -A -t -h  vcsesx001 -p 5432  -c "select * from pg_stat_replication"
  #  40576|10|postgres|walreceiver|10.209.70.89||58926|2018-09-10 17:37:17.436665+00||streaming|0/30467C0|0/30467C0|0/30467C0|0/30467C0||||1|sync
  #
  #-------------------------------------------------------------

  $sCommandArgs = "-A -t -h $sHostName -p $iPort -c \"select * from pg_stat_replication\"";

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  $iTimeout = 10;
  if ( PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser, 
                                     $sEnvFile, 
                                     $sCommand, 
                                     $sCommandArgs, 
                                     \$iExitCode, 
                                     $iTimeout, 
                                     \@lsCmdout ) ) {
    &logTrace ( 20335,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
    
    if ( 0 == $iExitCode )
    {
		&logTrace ( 20336,
		"<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
		$sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
      if ( ( $lsCmdout[0]) && ( grep(/\|streaming/,@lsCmdout) ) ) {
        &logTrace ( 20337,
                "<%s> Replication active.",
                $sMyName );
      }
      else {
        &logWarn ( 20338,
                "<%s> Replication NOT active.",
                $sMyName );
        $iReturnCode = 0;
      }
    }
    else 
    {
      &logWarn ( 20339,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]", 
      $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
      $iReturnCode = 0; # Failure..
    }
  }
  else {
    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}


=pod


 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String PostgreSQLUser,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $sEnvFile                  Absolute path to environment file to source
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPostgreSQLUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20079, "Subroutine <%s> called with:
  PostgreSQLUser           [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName,   $sPostgreSQLUser,   $sEnvFile,   $sCmd,    $sCmdArgs,
  $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20080,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20081,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20082,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20083,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20084,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser, 
                             $sEnvFile, 
                             $sCmd, 
                             $sCmdArgs, 
                             \$iExitCode, 
                             $iTimeout, 
                             \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20085,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20086,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20087, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20088, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20089,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20090,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20091,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}

=pod

-------------------------------------------------------------------------
Function : bool DeletePidFile ( );


Description : Deletes PID files for the VCS resource.

Parameters :

$sResourceName ResourceName for which the PID files needs to be deleted.

Returns : boolean value in the set of {0,1} where
0 false
1 true
PostgreSQL::DeletePidFile ( );
-------------------------------------------------------------------------

=cut
sub DeletePidFile {

#-----------------
# Announce Entry..
#-----------------
  my $sMyName = ( caller 0 )[3];
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];

  &logTrace ( 20317, "Subroutine <%s> called with:
      ResourceName        [%s]
      ResourceType        [%s]", 
      $sMyName, $sResourceName, $sResourceType );

#--------------------------------------
# Set Return Code Optimistically to 1..
#--------------------------------------
  my $iReturnCode = 1;

  my ( $sPidFileDir, $iRet ) = ();
#-------------------
# Delete PID files..
#-------------------
  &logTrace ( 20318,
      "<%s> Deleting PID files for VCS resource [%s]",
      $sMyName, $sResourceName );

  $sPidFileDir = "$VCSLOG/log/tmp/.VRTS$sResourceType/$sResourceName/PID";

  if ( -d $sPidFileDir ) {
    $iRet = system("/bin/rm -rf $sPidFileDir") ;

    if ( $iRet == 0 ) {
      &logTrace ( 20319,
          "<%s> Successfully deleted PID files in the directory [%s].",
          $sMyName, $sPidFileDir );
      $iReturnCode = 0;
    }
    else {
      &logWarn ( 20320,
          "<%s> Failed to delete PID files in directory [%s]. Remove the files manually.",
          $sMyName, $sPidFileDir );
      $iReturnCode = 1;
    }
  }
  else {
    &logTrace ( 20321,
        "<%s> PID files for VCS resource [%s] doesn't exist.",
        $sMyName, $sResourceName );
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20092, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20093, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20094, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20095, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20096,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20097,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20098,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( PostgreSQL::GetPidDetails ( $iPid,
                                 \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20099,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20100,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20101,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20102,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20103, "<%s> PID [%s] does not exist",
    $sMyName, $iPid );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillPostgreSQLProcesses ( IN String BaseDir,
                                           IN String DataDir,
                                           IN String HostName,
                                           IN String PostgreSQLUser,
                                           IN Boolean MonitorReplication,
                                           IN String ClientAddr );

 Description  : Kills processes for this instance of PostgreSQL
 
 Parameters :

 $sBaseDir                The Installation Directory of PostgreSQL.
 $sDataDir                Location of the database storage area.
 $sHostName               Hostname/IPv4 address of virtual host for PostgreSQL Instance
 $sPostgreSQLUser         Specifies the UserName required to run into pg_ctl.
 $bMonitorReplication       When enabled, Replication link is monitored. Takeover operation is enabled. Instance is connected as standby if Master is running.
 $sClientAddr               Address of the Standby instance
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillPostgreSQLProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDataDir,
       $sEnvFile,
       $sHostName,
       $iPort,
       $sPostgreSQLUser,
       $bMonitorReplication,
       $sClientAddr ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20104, "Subroutine <%s> called with:
  BaseDir                  [%s]
  DataDir                  [%s]
  HostName                 [%s]
  PostgreSQLUser           [%s]
  MonitorReplication       [%s]
  ClientAddr               [%s]",

  $sMyName, $sBaseDir, $sDataDir, $sHostName, $sPostgreSQLUser,
  $bMonitorReplication, $sClientAddr );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'PostgreSQL::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    PostgreSQL::GetProcState ( \@liGoners,
                               $sBaseDir,
                               $sDataDir,
                               $sEnvFile,
                               $sHostName,
                               $iPort,
                               $sPostgreSQLUser,
                               $bMonitorReplication,
                               $sClientAddr,
                               0 ); # Set filtering choice..
  }
  PostgreSQL::DeletePidFile ( ); 
  #--------------------------
  # Sort in ascending order..
  #--------------------------
  @liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 2, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20105,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20106,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
    sleep(2);
    my @liUnDead9 = ();
    if ( @liUnDead9 = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liUnDead ) {       # From what's found..
      &logError ( 20310,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead9 ), scalar @liUnDead9 );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20311,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 1;
    }

  }
  else {
    

    &logTrace ( 20107,
    "<%s> Could not locate parent process for this instance of PostgreSQL.",
    $sMyName );
  }
   
    #--------------------------------------------
    # If the parent process was killed
    # externally, there are chances that
    # the child processes still exist.
    # These child processes will now have ppid 1
    # and will start with "postgres:" pattern.
    # We will atempt to kill these processes
    #-------------------------------------------
    
    
    my ( $sLang, @hProcListHash, $iProcessCount, @liOnlineChild, @liUnDeadChild, $rhlProc, $iPid, $iUid, $sProc ) = ();

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iUid = getpwnam $sPostgreSQLUser;
    $sProc = "postgres:";
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );
    
    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20312,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20313,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return 0;
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      next if $$rhlProc { PPID } != 1;
      $iPid = $$rhlProc { PID };             # Assign..

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..

         if ( $$rhlProc { COMMAND } =~ /^$sProc.*/ ) {    # Now found..
             push ( @liOnlineChild, $iPid );         # Store..
             &logTrace ( 20314,
             "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
             $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
             $sProc );

        }
      }
   }

    @liOnlineChild = sort { $a <=> $b } @liOnlineChild;

   #------------------------------------------------------------------
   # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
   # valid process before we actually issue the SIGKILL. We don't want
   # to report a failure for SIGKILL in case  of  defunct  or  missing
   # processes.
   #------------------------------------------------------------------
   if ( scalar @liOnlineChild ) {
     #----------------------
     # Failed to kill some..
     #----------------------
     if ( @liUnDeadChild = grep {
                          kill ( 0, $_ ) && # Valid..
                        ! kill ( 9, $_ )    # and could NOT be killed..
                      } @liOnlineChild ) {       # From what's found..
       &logError ( 20315,
       "<%s> PIDs not killed [%s] Total not killed [%s]",
       $sMyName, join ( ' ', @liUnDeadChild ), scalar @liUnDeadChild );
       $iReturnCode = 0; # Failure..
     }
     else { # All killed..
       &logInfo ( 20316,
       "<%s> Killed PIDs [%s] Total killed: [%s]",
       $sMyName, join ( ' ', @liOnlineChild ), scalar @liOnlineChild );
     }
   }

  #------------------------------------
  # postmaster.pid file should not exist..
  #------------------------------------a
  my $sLockFile = "$sDataDir/postmaster.pid";
  if ( -f $sLockFile ) {
    if ( ! Sys::UnlinkFilesAsUser ( $sLockFile ) ) {
      &logError ( 20340,
          "<%s> Could not remove file [%s]", $sMyName, $sLockFile );
      return 0; # Cannot proceed..
    }
    &logTrace ( 20495,
    "<%s> Removed the Lock File [%s]", $sMyName, $sLockFile );
  }
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20108, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20109,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20110,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20111,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..
  #--------------------------------------------------------------
  # Triggers are executed from /var/VRTSvcs/diag/had
  # Postgres user does not have access to this directory,
  # chdir to PG home to avoid warnings while executing
  # psql commands.
  #--------------------------------------------------------------
  if ( ! defined $sUser || $sUser eq "" ) {
    $sUser = "root";
  }
  my $sFromDir = (getpwnam($sUser))[7];
  &logTrace ( 20112,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout,
                                                           "",
                                                           $sFromDir );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20113,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20114,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20115,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20116,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName : delete $ENV { LOGNAME };

 $$rlsCmdout [0] ||= ''; # Set if not defined..

  return $iReturnCode;
}



=pod
 -------------------------------------------------------------------------
 Function : bool IsMasterRunning ( IN String BaseDir,
                                   IN String PostgreSQLUser,
                                   IN String EnvFile,
                                   IN String SourceConnStr,
                                   IN Integer TillWhen );



 Description : Checks if Master is running.
               Uses Postgres command "select pg_is_in_recovery()"

 Parameters :
 
 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $sEnvFile                  Absolute path to environment file to source
 $sSourceConnStr            Connection string used by pg_rewind command
 $iTillWhen                 Time in EPOCH left for the operation

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut


sub IsMasterRunning {

  #-----------------
  # Announce Entry..
  #-----------------
  my ( $sBaseDir, 
       $sPostgreSQLUser,
       $sEnvFile,
       $sSourceConnStr,
       $iTillWhen,
       $sMasterNode,
       $iPort ) = @_;
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20466, "Subroutine <%s> called with:
  BaseDir                     [%s]
  PostgreSQLUser              [%s]
  EnvFile                     [%s]
  SourceConnStr               [%s]
  TillWhen                    [%s]
  Master                      [%s]
  Port                        [%s]",
  $sMyName, $sBaseDir, $sPostgreSQLUser, $sEnvFile, $sSourceConnStr, $iTillWhen, $sMasterNode, $iPort );


  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my ( $iExitCode, $sCommand, $sCommandArgs, $sGroupName, $sGlobalGrp, $sTargetSys ) = ();
  my (@lsOut, $sOut) = ();

  my ( $iTimeout, @lsCmdout ) = ();
  my ( $sReplicationModeTuples, $sResourceName, $sPgGroup, @lsCmdOut, $sRepGrp, $sLine, $sMaster ) = ();
  
  $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  if ( PostgreSQL::GetAttrValue("ReplicationModeTuples", $sResourceName,"", \$sReplicationModeTuples ) && defined( $sReplicationModeTuples )  && $sReplicationModeTuples ne "" ) {
    $iExitCode = PGReplication::GetMaster( $sBaseDir, $iPort, \$sMaster, $sPostgreSQLUser, $sEnvFile, $iTillWhen );
    if ( $iExitCode == 1 ) {
      $$sMasterNode = $sMaster;
      return 1;
    }
    elsif ( $iExitCode == 2 ) {
      return 2;#Should not start Node
    }
    else {
      #master not found
      return 0;
    }
  }
  else {
    my $sMasterHost = ($sSourceConnStr =~ m/host=(.*?)[\s|\$]/g)? $1 : undef ;
    my $sMasterPort = ($sSourceConnStr =~ m/port=(.*?)[\s|\$]/g)? $1 : undef;


    if (!defined $sMasterHost) {
      &logError ( 20474,
      "<%s> MasterHost [%s] not defined in SourceConnStr[%s]",
      $sMyName,  $sMasterHost, $sSourceConnStr);
      return 0; # Failure..    
    } 
    $sMasterPort = 5432 unless defined $sMasterPort;
    #---------------------------------------------
    # Declare monitor command and args variables..
    #---------------------------------------------
    $sCommand = "$sBaseDir/psql";
    $sCommandArgs = "-A -x -h $sMasterHost -p $sMasterPort -c \"select pg_is_in_recovery\(\)\"";
    $iTimeout = $iTillWhen - $^T;
    if ( PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser,
                                    $sEnvFile,
                                    $sCommand,
                                    $sCommandArgs,
                                    \$iExitCode,
                                    $iTimeout,
                                    \@lsCmdout ) ) {
      &logTrace ( 20330,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );

    if ( 0 == $iExitCode )
    {
                &logTrace ( 20331,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
      chomp $lsCmdout[0];
      if ( grep(/pg_is_in_recovery\|t/,@lsCmdout) ) {
        &logInfo ( 20332,
                "<%s> Standby server is running. Starting as Master. ",
                $sMyName );
        return 0;
      }
      elsif ( grep(/pg_is_in_recovery\|f/,@lsCmdout) ) {
        &logInfo ( 20452,
                "<%s> Master server is running. ",
                $sMyName );
        return 1;
      }
      else {
        &logInfo ( 20515,
                "<%s> pg_is_in_recovery did not return true or false.",
                $sMyName );
      }
    }
    else
    {
      &logError ( 20333,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
       return 0 # Failure..
    }
  }
  else
  {
    &logWarn ( 20334,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }
 }
}


=pod

 -------------------------------------------------------------------------
 Function : int ConnectAsStandby ( IN String BaseDir,
                                   IN String DataDir,
                                   IN Integer Port,
                                   IN String HostName,
                                   IN String PostgreSQLUser,
                                   IN String SourceConnStr,
                                   IN String BackupCmd,
                                   IN String RecoveryFile,
                                   IN String EnvFile,
                                   IN Integer TillWhen ); 
          


 Description : Runs Promote Action Entry Point if current mode is standby
               and the EntryPoint is called as a result of 
               switchover or failover.

 Parameters :
 
 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sDataDir                  Location of the database storage area.
 $iPort                     The port number on which PostgreSQL instance is configured.
 $sHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $sSourceConnStr            Connection string used by pg_rewind command
 $sBackupCmd                Command used to base backup PostgreSQL data directory
 $sRecoveryFile             recovery.conf file provided to start standby instance
 $sEnvFile                  Absolute path to environment file to source
 $iTillWhen                 Time in EPOCH left for the operation

 Returns : integer value in the set of {-1,0,1} where
          -1 Rewind failed
           0 false
           1 true
 -------------------------------------------------------------------------

=cut


sub ConnectAsStandby {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir, 
       $sDataDir, 
       $iPort, 
       $sHostName, 
       $sPostgreSQLUser, 
       $sSourceConnStr, 
       $sBackupCmd, 
       $sRecoveryFile,
       $sEnvFile, 
       $iTillWhen,
       $sMaster  ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20385, "Subroutine <%s> called with:
  BaseDir             [%s]
  DataDir             [%s]
  Port                [%s]
  HostName            [%s]
  PostgreSQLUser      [%s]
  SourceConnStr       [%s]
  BackupCmd           [%s]
  RecoveryFile        [%s]
  EnvFile             [%s]
  TillWhen            [%s]
  Master              [%s]",
   
  $sMyName, $sBaseDir, $sDataDir, $iPort, $sHostName, 
  $sPostgreSQLUser, $sSourceConnStr, $sBackupCmd, $sRecoveryFile,
  $sEnvFile, $iTillWhen, $sMaster );


  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  my ( $iExitCode, $ret, $sCommand, $sCommandArgs, $sReplicationModeTuples ) = ();
  my (@lsOut, $sOut) = ();
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  opendir(my $dh, $sDataDir);
  if ( scalar(grep { $_ ne "." && $_ ne ".." } readdir($dh)) == 0) { # If data directory is empty
    if ($ret = PGReplication::BaseBackup( $sDataDir, 
                                          $sPostgreSQLUser, 
                                          $sBackupCmd, 
                                          $sRecoveryFile,
                                          $sEnvFile, 
                                          $iTillWhen )){
      &logTrace ( 20386,
                  "<%s> Base Backup successful.",
                  $sMyName);

    }
    else {
      &logTrace ( 20387,
                 "<%s> Base Backup failed.",
                  $sMyName);
      $iReturnCode = 0;

    }
  }
  else {
    if ( PostgreSQL::GetAttrValue("ReplicationModeTuples", $sResourceName,"", \$sReplicationModeTuples ) && defined( $sReplicationModeTuples ) && $sReplicationModeTuples ne "" ) {
      $ret = PGReplication::Rewind( $sBaseDir, 
                                    $sDataDir, 
                                    $iPort, 
                                    $sHostName, 
                                    $sPostgreSQLUser, 
                                    $sSourceConnStr, 
                                    $sRecoveryFile,
                                    $sEnvFile, 
                                    $iTillWhen);
      if ( $ret == 1 ) {
        &logTrace ( 20390,
        "<%s> rewind successful.",
       $sMyName);
      }
      elsif ( $ret == 2 ) {
        return 2;#GetMaster failed with error
      }
      else {
        &logTrace ( 20391,
        "<%s> rewind failed",
        $sMyName);
        $iReturnCode = -1;
      }
    }
    else {
    if ( PostgreSQL::IsRecoveryFileFound($sDataDir,
                                         $sRecoveryFile ) ){
      &logTrace ( 20389,
                  "<%s> Recovery file found. Rewind not required.",
                   $sMyName, $sRecoveryFile);       
      if ($sRecoveryFile !~ /recovery.conf/) {
        my $sCommand = "$sCmdTouch";
        my $sCommandArgs = "$sDataDir/standby.signal";
        my $iTimeout = $iTillWhen - time;
        my ( @lsCmdout, $iExitCode ) = ();
        
        if ( PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser,
                                        "",
                                        $sCommand,
                                        $sCommandArgs,
                                        \$iExitCode,
                                        $iTimeout,
                                        \@lsCmdout ) ) {
          &logTrace ( 20610,
          "<%s> Cmd [%s] with args [%s] executed successfully",
          $sMyName, $sCommand, $sCommandArgs );
      
          if ( 0 == $iExitCode )
          {
                      &logTrace ( 20611,
                      "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                      $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
          }
          else
          {
            &logError ( 20612,
            "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
             $sMyName, $iExitCode, join ( chr 012, @lsCmdout ) );
             return 0 # Failure..
          }
        }
        else
        {
          &logError ( 20613,
          "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
          $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
          return 0; # Failure..
        }
      }
         ( my $sFile = $sRecoveryFile ) =~ s#.*/(.*)$#$1#;
         Sys::UnlinkFilesAsUser ( "$sDataDir/$sFile" );
       my $iExitCode = system ("$sCmdCP -p $sRecoveryFile $sDataDir");
       if ( $iExitCode == 0 ) {
           &logTrace ( 20606,
              "<%s> Cmd [%s] executed successfully",
              $sMyName, $sCmdCP);
        }
        else {
           &logError ( 20607,
              "<%s> Cmd [%s] failed to execute",
              $sMyName, $sCmdCP);
        return 0; 
        } 
     }
    elsif ($ret = PGReplication::Rewind($sBaseDir, 
                                        $sDataDir, 
                                        $iPort, 
                                        $sHostName, 
                                        $sPostgreSQLUser, 
                                        $sSourceConnStr, 
                                        $sRecoveryFile,
                                        $sEnvFile, 
                                        $iTillWhen)) {
       &logTrace ( 20390,
                  "<%s> rewind successful.",
                   $sMyName);
    }
    else {
       &logTrace ( 20391,
                  "<%s> rewind failed",
                   $sMyName);
       $iReturnCode = -1;
    }
  }
 }
return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20117, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20118,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  #goto $sSwitch; {

  if ( "solaris" eq $sSwitch ) {
      my $sOSVer = ( uname ) [2]; # Get OS version..
      if ($sOSVer =~ /\d+\.(\d+)/) {
          $sOSVer = $1;
      }
      $sPsCmd = "/usr/bin/ps";
      if ( $sOSVer < 11 ) {
          $sPsCmd = "/usr/ucb/ps" if ( -x "/usr/ucb/ps" );
      }

      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  }
  elsif ( "hpux" eq $sSwitch ) {
      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }
  elsif ( "aix" eq $sSwitch ) {
      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  }
  elsif ( "linux" eq $sSwitch ) {
      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }

  &logTrace ( 20119, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20120,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95 : delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20121,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );
    PostgreSQL::DeletePidFile ( );
    %hsProc = (); # Truncate..

    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String BaseDir,
                                  IN String DataDir,
                                  IN String HostName,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sDataDir                  Location of the database storage area.
 $sHostName                 Hostname/IPv4 address of virtual host for PostgreSQL Instance
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDataDir,
       $sHostName,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20122, "Subroutine <%s> called with:
  BaseDir                  [%s]
  DataDir                  [%s]
  HostName                 [%s]
  SrvrDetails              [%s]",

  $sMyName, $sBaseDir, $sDataDir, $sHostName, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $iPat ) = ();

  #-------------------------------------
  # List of search patterns to monitor..
  #-------------------------------------
  my @lsSrchPatterns = ( qr@($sBaseDir/|^)(edb-|)post(gres|master)\b.* -D $sDataDir\b.*@ );

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "$VCSLOG/log/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20123, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20124, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  foreach $sProc ( @lsSrchPatterns ) {
    $sKey = 'srchstr_' . ++ $iPat;
    $sPidFile = "$sPIDDir/pidfile.$sKey";
    @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20125, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;

  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20126,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
  }
  else {
    if ( lc $sEntryPointName ne "preonline" ) {
      #-----------------------------------------
      # Form query string based on entry point..
      #-----------------------------------------
      $sCommand = $sCmdHares; # Get global..
      $sCommandArgs = "-value $sResourceName $sAttribute";
    
      #----------------------------------
      # Fire command and collect output..
      #----------------------------------
      ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                        =~ s/\n//;
    
      #--------------------
      # Check exit status..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logTrace ( 20127,
        "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode );
      }
      elsif ( ! $iEngineTimeout ) { # Undefined value..
        &logTrace ( 20128, "<%s> Could not get [%s] from engine",
        $sMyName, $sAttribute );
      }
      else { # Success..
        &logTrace ( 20129, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
      }
    }
    
      #-----------------------------------------
      # Form query string based on entry point..
      #-----------------------------------------
    if ( lc $sEntryPointName eq "preonline" ) {
     $sEntryPointName = "offline" ;
    }
    unless ( $iEngineTimeout ) {
      $sCommand = $sCmdHatype; # Get global..
      ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
      $sResourceType = "PostgreSQL" if $sResourceType !~ /\w/; # Ensure..
      $sCommandArgs = "-value $sResourceType $sAttribute";
    
      #----------------------------------
      # Fire command and collect output..
      #----------------------------------
      ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
    
      #--------------------
      # Check exit status..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logTrace ( 20130,
        "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      elsif ( ! $iEngineTimeout ) { # Empty value..
        &logTrace ( 20131, "<%s> Could not get [%s] from engine",
        $sMyName, $sAttribute );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      else { # Success..
        &logTrace ( 20132, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
      }
    }
  }

  if ( lc $sEntryPointName eq "preonline" ) {
    $iEngineTimeout *= 2;
    &logTrace ( 20583,
    "<%s> This is the preonline entrypoint. Timeout value is set to twice the offline timeout. Preonline timeout value is [%s]",
    $sMyName, $iEngineTimeout );
  }
  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20133,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}


=pod
--------------------------------------------------------------------------------
 Function: bool GetNodeId( OUT Reference NodeId )

 Description: This subroutine returns VCS nodeid of host where agent is running

 Parameters:
 NodeId            Node name of current host 

 Returns : boolean value in the set of {0,1} where
           0 Successfully retrieved node name.
           1 Unsuccessful.
--------------------------------------------------------------------------------
=cut

sub GetNodeId{
  
  my $sCurrentHost = shift;
  my $sMyName = ( caller 0 )[3];
  &logTrace( 20460, "Subroutine <%s> called:
                     CurrentHost [%s]", $sMyName, $sCurrentHost );
  my $iExitCode = 0;
  my $sCommand = $sCmdHasys;
  my $sCommandArgs = " -nodeid ";
  $$sCurrentHost = readpipe ( "$sCommand $sCommandArgs" );
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20461, "<%s>: Cmd [%s] with args [%s] failed with exit code [%s].", $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    return 1;
  }

  chomp $$sCurrentHost;
  $$sCurrentHost =~ s/.*Node name: //;
  $$sCurrentHost =~ s/ Node ID: .*//;
  &logTrace( 20462, "<%s>: Current host is [%s]", $sMyName, $$sCurrentHost );

  return 0;
}


=pod
--------------------------------------------------------------------------------
 Function: bool RemoteGroupOperation( IN String Group,
                                      IN String Operation,
                                      IN String System,
                                      IN String ExpectedState,
                                      IN Integer TillWhen )

 Description: This subroutine performs group operation on the remote postgreSQL group

 Parameters:
 Group            Name of the Service group on which the operation is to be performed
 Operation        Name of the Operation (online/offline/clear)
 System           Name of the remote system where the operation is to be performed
 ExpectedState    Expected State of the Resource after command
 TillWhen         Time in EPOCH left for Operation   

 Returns : boolean value in the set of {0,1} where
           0 True
           1 False
--------------------------------------------------------------------------------
=cut

sub RemoteGroupOperation{
  
  #----------------
  # Get Arguments..
  #----------------
  my ( $sGroup,
       $sOperation,
       $sSystem,
       $sExpectedState,
       $iTillWhen ) = @_;

  my $sMyName = ( caller 0 )[3];
  &logTrace( 20475, "Subroutine <%s> called with :
  Group              [%s]
  Operation          [%s]
  System             [%s]
  ExpectedState      [%s]
  TillWhen           [%s]", 
  $sMyName, $sGroup, $sOperation, $sSystem, $sExpectedState, $iTillWhen );
  my ( $iTimeout, $sCommand, $sCommandArgs, $iSleepTime, @lsCmdout )  = ();
  my $iExitCode = 0;
  

  $sCommand = $sCmdSSH;
  #---------------------------------------------------------------------
  # #ssh vmrac018 '/opt/VRTSvcs/bin/hagrp -offline PG_SG -sys vmrac018'
  #----------------------------------------------------------------------
  $sCommandArgs = " $sSystem \'$HAGRP -$sOperation $sGroup -sys $sSystem\'";
  &logTrace ( 20476, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
  $iTimeout = $iTillWhen - time;
  if ( PostgreSQL::ExecLocalCmd ( "",
                                  "",
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20478,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( 0 == $iExitCode )
    {
                &logTrace ( 20508,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
    }
    else
    {
      &logError ( 20477,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, join ( chr 012, @lsCmdout ) );
       return 0 # Failure..
    }
  }
  else
  {
    &logError ( 20509,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }

#Delay operation if delay enabled

    return 1 unless ($sExpectedState ne "") ;
   #-----------------
   # Set sleep time..
   #-----------------
   $iSleepTime = ($sOperation eq "clear")? 5 : 10; # Sec..
   
 
   &logInfo ( 20479,
   "<%s> Delaying exit to confirm operation[%s]..", $sMyName,$sOperation );

   #--------------------
   # Delay entry point..
   #--------------------
   while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
     #-----------------------
     # Wait before retrying..
     #-----------------------
     sleep $iSleepTime; # Secs..
 
     $sCommand = $sCmdSSH;
     $sCommandArgs = " $sSystem \'$HAGRP -value $sGroup State\'";
     &logTrace ( 20480, "<%s> Executing Command [%s] with arguments [%s]",
                   $sMyName, $sCommand, $sCommandArgs );
     if ( PostgreSQL::ExecLocalCmd ( "",
                                     "",
                                     $sCommand,
                                     $sCommandArgs,
                                     \$iExitCode,
                                     $iTimeout,
                                     \@lsCmdout ) ) {
       &logTrace ( 20482,
       "<%s> Cmd [%s] with args [%s] executed successfully",
       $sMyName, $sCommand, $sCommandArgs );
   
       if ( 0 == $iExitCode )
       {
                   &logTrace ( 20510,
                   "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                   $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
          chomp @lsCmdout;
          if ($lsCmdout[0] eq "$sExpectedState"){
            &logInfo ( 20483, "<%s> Confirmed operation[%s] of group[%s]",
                       $sMyName, $sOperation, $sGroup );
            return 1;
          }
          else {
            &logTrace ( 20484, "<%s> Could not confirm operation[%s]. Retrying in [%s] sec.
                                Delaying with [%s] sec in hand",
                       $sMyName,  $sOperation, $iSleepTime, $iTimeout - $iSleepTime );
          }
       }
       else
       {
         &logError ( 20481,
         "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
         $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
          return 0 # Failure..
       }
     }
     else
     {
       &logError ( 20511,
       "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
       return 0; # Failure..
     }
   }
return 0;
}

=pod
--------------------------------------------------------------------------------
 Function: bool GetPGResName( IN  String $sPGGroupName,
                              OUT Reference $sPGResourceName,
                              IN  Integer TillWhen )

 Description: This subroutine returns PostgreSQL resource name from group

 Parameters:
 PGGroupName               Name of PostgreSQL group
 PGResourceName            Name of PostgreSQL resource
 TillWhen                  Time in EPOCH left for Operation

 Returns : boolean value in the set of {0,1} where
           0 Successfully retrieved the ResourceName.
           1 Unsuccessful.
--------------------------------------------------------------------------------
=cut

sub GetPGResName{
    my ( $sPGGroupName,
         $sPGResourceName,
         $iTillWhen ) = @_;

  my $sMyName = ( caller 0 )[3];
  &logTrace( 20494, "Subroutine <%s> called with :
  PGGroupName              [%s]
  PGResourceName           [%s]
  TillWhen                 [%s]",
  $sMyName, $sPGGroupName, $sPGResourceName, $iTillWhen );

  my ( @lsOut, @lsCmdout, $sCommand, $sCommandArgs ) = ();
  my $iExitCode = 0;
  my $iTimeout = $iTillWhen - $^T;


  #------------------------------------------------------
  # Find the PostgreSQL Resource. 
  # 
  # Sample output:
  #
  # #hares -list Type=PostgreSQL Group=PG_SG -localclus
  # pg_res                  vmrac018
  # pg_res                  vmrac019
  #
  # Note : When Replication is configured, agent does not 
  # support multiple PostgresSQL resources in the same service group
  #-------------------------------------------------------
  $sCommand = $sCmdHares;
  $sCommandArgs = "-list Type=PostgreSQL Group=$sPGGroupName -localclus";
  &logTrace ( 20442, "<%s> Executing Command [%s] with arguments [%s]",
              $sMyName, $sCommand, $sCommandArgs );
  if ( PostgreSQL::ExecLocalCmd ( "",
                                  "",
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20582,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( 0 == $iExitCode )
    {
                &logTrace ( 20512,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
  @lsOut = split( '\s+', $lsCmdout[0]);
  $$sPGResourceName = $lsOut[0];
  
  if (! (defined $$sPGResourceName) || ($$sPGResourceName =~ /^\s*$/)) { 
    &logError ( 20469,
      "<%s> Could not find the PostgreSQL ResourceName.",
      $sMyName);
    return 0;
  }
  &logTrace ( 20445,
      "<%s> Current PostgreSQL ResourceName [%s].",
      $sMyName, $$sPGResourceName );
  

    }
    else
    {
      &logError ( 20443,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
       return 0 # Failure..
    }
  }
  else
  {
    &logError ( 20513,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }
  
  
  return 1;
}


=pod
--------------------------------------------------------------------------------
 Function: bool GetAttrValue( IN  String $sAttributeName,
                              IN  String $sResourceName,
                              IN  String $sAgentType,
                              OUT Reference $sAttributeValue )

 Description: This subroutine returns PostgreSQL resource value

 Parameters:
 PGGroupName               Name of PostgreSQL group
 PGResourceName            Name of PostgreSQL resource

 Returns : boolean value in the set of {0,1} where
           0 Successfully retrieved the ResourceName.
           1 Unsuccessful.
--------------------------------------------------------------------------------
=cut

sub GetAttrValue{
    my ( $sAttributeName,
         $sResourceName,
         $sSystem,
         $sAttributeValue ) = @_;

  my $sMyName = ( caller 0 )[3];
  &logTrace( 20530, "Subroutine <%s> called with :
  AttributeName              [%s]
  ResourceName               [%s]
  System                     [%s]
  AttributeValue             [%s]",
  $sMyName, $sAttributeName, $sResourceName, $sSystem, $$sAttributeValue );

  my ( @lsHaresOut, @lsOut, $sCommand, $sCommandArgs ) = ();
  my $iExitCode = 0;


  #------------------------------------------------------
  # Find the PostgreSQL Attribute Value
  # 
  # Sample output:
  # #hares -value pg_res RegistrationOfStandby vcsesx001 -localclus
  #  0
  #-------------------------------------------------------
  $sCommand = $sCmdHares;
  $sCommandArgs = "-value $sResourceName $sAttributeName $sSystem -localclus";
  &logTrace ( 20531, "<%s> Executing Command [%s] with arguments [%s]",
              $sMyName, $sCommand, $sCommandArgs );
  ( $$sAttributeValue = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                               =~ s/\n//;

  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20532,
                "<%s> Cmd [%s] with args [%s] failed with exit code [%s].",
                $sMyName, $sCommand, $sCommandArgs, $iExitCode);
    return 0;
  }
  else {
    &logTrace ( 20533,
                "<%s> Cmd [%s] with args [%s] succeeded with exit code [%s].",
                $sMyName, $sCommand, $sCommandArgs, $iExitCode);
  }
  
#  if (! (defined $$sAttributeValue) || ($$sAttributeValue =~ /^\s*$/)) { 
  if (! (defined $$sAttributeValue) ) { 
    &logError ( 20534,
      "<%s> Could not find the [%s] Attribute value.",
      $sMyName, $sAttributeName);
    return 0;
  }
  &logTrace ( 20535,
      "<%s> [%s] Attribute value [%s].",
      $sMyName, $sAttributeName, $$sAttributeValue );
  
  return 1;
}

=pod
--------------------------------------------------------------------------------
 Function: bool GetSysState( IN String System
                             OUT Reference SysState )

 Description: This subroutine returns the status of System

 Parameters:
 Sys            System name
 SysState       Status of the System

 Returns : boolean value in the set of {0,1} where
           0 Successfully retrieved system state.
           1 Unsuccessful.
--------------------------------------------------------------------------------
=cut

sub GetSysState{
  
 my ( $sSystem,
      $sSysState ) = @_;

  my $sMyName = ( caller 0 )[3];
  &logTrace( 20573, "Subroutine <%s> called with :
  System              [%s]
  SysState            [%s]",
  $sMyName, $sSystem, $$sSysState );
  my $iExitCode = 0;
  my $sCommand = $sCmdHasys;
  my $sCommandArgs = " -state $sSystem ";
  &logTrace ( 20574, "<%s> Executing Command [%s] with arguments [%s]",
              $sMyName, $sCommand, $sCommandArgs );
  ( $$sSysState = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                               =~ s/\n//;
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20575, "<%s>: Cmd [%s] with args [%s] failed with exit code [%s].", $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    return 1;
  }

  chomp $$sSysState;
  &logTrace( 20576, "<%s>: System state for System[%s] is [%s]", $sMyName, $sSystem, $$sSysState );
  return 0;
}

=pod
--------------------------------------------------------------------------------
 Function: bool CheckSSH ( IN String System 
                           IN String TillWhen)

 Description: This subroutine checks if the agent is able to SSH in the system

 Parameters:
 System            System name
 TillWhen          Time in EPOCH left for command execution

 Returns : boolean value in the set of {0,1} where
           0 SSH unsuccessful.
           1 SSH successful.
--------------------------------------------------------------------------------
=cut

sub CheckSSH {
  
 my ( $sSystem, $iTillWhen ) = @_;

  my $sMyName = ( caller 0 )[3];
  &logTrace( 20594, "Subroutine <%s> called with :
  System              [%s]
  TillWhen            [%s]",
  $sMyName, $sSystem, $iTillWhen );
  my ($iExitCode, @lsCmdout) = ();
  my $sCommand = $sCmdSSH;
  my $sCommandArgs = " -q $sSystem exit";
  &logTrace ( 20595, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
  my $iTimeout = $iTillWhen - time;
  if ( PostgreSQL::ExecLocalCmd ( "",
                                  "",
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20596,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( 0 == $iExitCode )
    {
                &logTrace ( 20597,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
    }
    else
    {
      &logError ( 20598,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
       return 0 # Failure..
    }
  }
  else
  {
    &logError ( 20599,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }
  return 1;
}


=pod
 -------------------------------------------------------------------------
 Function : bool IsRecoveryFileFound($sDataDir,
                                     $sRecoveryFile ) 
 -------------------------------------------------------------------------
=cut

sub IsRecoveryFileFound {
  #----------------
  # Get Arguments..
  #----------------
  my($sDataDir, $sRecoveryFile) = @_;
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  
  &logTrace ( 20614, "Subroutine <%s> called with
  DataDir            [%s]
  RecoveryFile       [%s]",
  $sMyName, $sDataDir, $sRecoveryFile);

  #-------------------------
  # Define local variables..
  #-------------------------
  ( my $sFile = $sRecoveryFile ) =~ s#.*/(.*)$#$1#;
  $sFile = "standby.signal" unless ($sFile eq "recovery.conf"); 
  if (-e "$sDataDir/$sFile"){
    &logTrace ( 20615, "<%s>: Recovery File [%s] found in data directory [%s]",
            $sMyName, $sFile, $sDataDir );
    return 1;
  }
  else {
     &logTrace ( 20616, "<%s>: Recovery File [%s] not found in data directory [%s]",
               $sMyName, $sFile, $sDataDir );
    return 0;
  }
}
=pod
 -------------------------------------------------------------------------
 Function : bool IsMasterInRF( IN String RecoveryFile,
                               IN String Master ) 
 
 Description: This subroutine verify the entry of master server from recovery file.

 Parameters:

 $sRecoveryFile       recovery.conf file provided to start standby instance
 $sMaster             The node name on which master server is running.
 
 Returns : boolean value in the set of {0,1} where
           0 true.
           1 false.
 -------------------------------------------------------------------------
=cut 

sub IsMasterInRF {
  my ( $sRecoveryFile, $sMaster ) = @_;
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  
  &logTrace ( 20614, "Subroutine <%s> called with
  RecoveryFile       [%s]
  Master             [%s]",
  $sMyName, $sRecoveryFile, $sMaster );

  my $data;
  if ( open my $in, '<:encoding(UTF-8)', $sRecoveryFile ) {
    local $/ = undef;
    my $all = <$in>;
    close $in;
    ($all =~ m/host=((\w+|-|\.)*)/);
    my $sTmp = $1;
    if ( $sTmp eq $sMaster ) {
      &logTrace ( 20616,
      "<%s> Master [%s] is found in recovery file [%s]",
      $sMyName, $sMaster, $sRecoveryFile );
      return 1;
    } 
    else {
      &logTrace ( 20616, 
      "<%s> Master [%s] is not found in recovery file [%s]",
      $sMyName, $sMaster, $sRecoveryFile);
      return 0;
     }
  }
}

1; # So PERL knows its included..
